@@ -1,5 +1,5 @@
 /* OpenCP Module Player
- * copyright (c) 2020-'22 Stian Skjelstad <stian.skjelstad@gmail.com>
+ * copyright (c) 2020-'26 Stian Skjelstad <stian.skjelstad@gmail.com>
  *
  * Glue logic for using libsidplayfp playback engine
  *
@@ -18,8 +18,6 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-
-
 #include "libsidplayfp-api.h"
 #include <iostream>
 #include <fstream>
@@ -28,21 +26,32 @@
 #include "sidplayfp/SidInfo.h"
 #include "sidplayfp/SidTuneInfo.h"
 #include "builders/residfp-builder/residfp.h"
-#include "builders/resid-builder/resid.h"
+#include "builders/residfpII-builder/residfpII.h"
 
 extern "C"
 {
-#include "config.h"
+#ifdef DPACKAGE_NAME
+#undef DPACKAGE_NAME
+#endif
+
+#ifdef VERSION
+#undef VERSION
+#endif
+
+#ifdef PACKAGE_VERSION
+#undef PACKAGE_VERSION
+#endif
+
+#include "../config.h"
+#include <stdio.h>
+#include <unistd.h>
 #include "types.h"
 #include "boot/psetting.h"
 #include "filesel/dirdb.h"
 #include "filesel/filesystem-drive.h"
 #include "filesel/filesystem.h"
-#include <stdio.h>
 }
 
-#include <unistd.h>
-
 namespace libsidplayfp
 {
 #if 0
@@ -69,15 +78,15 @@ namespace libsidplayfp
 	{
 		switch (model)
 		{
-			case MOS656X::MOS6567R56A:
+			case MOS656X::model_t::MOS6567R56A:
 				return "MOS6567R56A NTSC-M (old)";
-			case MOS656X::MOS6567R8:
+			case MOS656X::model_t::MOS6567R8:
 				return "MOS6567R8 NTSC-M";
-			case MOS656X::MOS6569:
+			case MOS656X::model_t::MOS6569:
 				return "MOS6569 PAL-B";
-			case MOS656X::MOS6572:
+			case MOS656X::model_t::MOS6572:
 				return "MOS6572 PAL-N";
-			case MOS656X::MOS6573:
+			case MOS656X::model_t::MOS6573:
 				return "MOS6573 PAL-M";
 			default:
 				return "MOS65xx ??";
@@ -236,6 +245,16 @@ namespace libsidplayfp
 			//fprintf (stderr, "filtercurve6581=%lf\n", m_filter.filterCurve6581);
 		}
 
+		const char *range6581 = configAPI->GetProfileString("libsidplayfp", "filterrange6581", "0.5");
+		m_filter.filterRange6581 = strtod(range6581, &endptr);
+		if ((*endptr != 0) || (range6581 == endptr))
+		{
+			fprintf (stderr, "[libsidplayfp]\n  filterrange6581=invalid... defaulting to 0.5\n");
+			m_filter.filterRange6581=0.5;
+		} else {
+			//fprintf (stderr, "filterrange6581=%lf\n", m_filter.filterRange6581);
+		}
+
 		const char *curve8580 = configAPI->GetProfileString("libsidplayfp", "filtercurve8580", "0.5");
 		m_filter.filterCurve8580 = strtod(curve8580, &endptr);
 		if ((*endptr != 0) || (curve8580 == endptr))
@@ -246,31 +265,46 @@ namespace libsidplayfp
 			//fprintf (stderr, "filtercurve8580=%lf\n", m_filter.filterCurve8580);
 		}
 
+		const char *CWS = configAPI->GetProfileString("libsidplayfp", "combinedwaveforms", "Average");
+		if (!strcasecmp(CWS, "Weak"))
+		{
+			m_filter.combinedWaveforms = SidConfig::WEAK;
+		} else if (!strcasecmp(CWS, "Strong"))
+		{
+			m_filter.combinedWaveforms = SidConfig::STRONG;
+		} else if (!strcasecmp(CWS, "Average"))
+		{
+			m_filter.combinedWaveforms = SidConfig::AVERAGE;
+		} else {
+			fprintf (stderr, "[libsidplayfp]\n  combinedwaveforms=invalid... defaulting to Average\n");
+			m_filter.combinedWaveforms = SidConfig::AVERAGE;
+		}
+
 		m_engCfg.digiBoost = configAPI->GetProfileBool("libsidplayfp", "digiboost", 0, 0);
 		//fprintf (stderr, "digiboost=%d\n", m_engCfg.digiBoost);
 
-		// TODO, add check of this return value
-		createSidEmu (configAPI);
+		if (!createSidEmu (configAPI))
+		{
+			#warning handle this failure
+		}
 
 		const char *kernal_string  = configAPI->GetProfileString("libsidplayfp", "kernal",  "KERNAL.ROM");
 		const char *basic_string   = configAPI->GetProfileString("libsidplayfp", "basic",   "BASIC.ROM");
 		const char *chargen_string = configAPI->GetProfileString("libsidplayfp", "chargen", "CHARGEN.ROM");
 
-		uint32_t dirdb_base = dirdbAPI->ResolvePathWithBaseAndRef (dmFile->basedir->dirdb_ref, configAPI->ConfigDir, 0, dirdb_use_dir); /* should be the parent_dir of the file you want to load */
+		uint32_t dirdb_base = configAPI->DataHomeDir->dirdb_ref; /* should be the parent_dir of the file you want to load */
 		uint32_t kernal_ref;
 		uint32_t basic_ref;
 		uint32_t chargen_ref;
 
-#ifdef __W32__
-		kernal_ref  = dirdbAPI->ResolvePathWithBaseAndRef (dirdb_base, kernal_string,  DIRDB_RESOLVE_DRIVE | DIRDB_RESOLVE_TILDE_HOME | DIRDB_RESOLVE_WINDOWS_SLASH);
-		basic_ref   = dirdbAPI->ResolvePathWithBaseAndRef (dirdb_base, basic_string,   DIRDB_RESOLVE_DRIVE | DIRDB_RESOLVE_TILDE_HOME | DIRDB_RESOLVE_WINDOWS_SLASH);
-		chargen_ref = dirdbAPI->ResolvePathWithBaseAndRef (dirdb_base, chargen_string, DIRDB_RESOLVE_DRIVE | DIRDB_RESOLVE_TILDE_HOME | DIRDB_RESOLVE_WINDOWS_SLASH);
-
+#ifdef _WIN32
+		kernal_ref  = dirdbAPI->ResolvePathWithBaseAndRef (dirdb_base, kernal_string,  DIRDB_RESOLVE_DRIVE | DIRDB_RESOLVE_TILDE_HOME | DIRDB_RESOLVE_WINDOWS_SLASH, dirdb_use_file);
+		basic_ref   = dirdbAPI->ResolvePathWithBaseAndRef (dirdb_base, basic_string,   DIRDB_RESOLVE_DRIVE | DIRDB_RESOLVE_TILDE_HOME | DIRDB_RESOLVE_WINDOWS_SLASH, dirdb_use_file);
+		chargen_ref = dirdbAPI->ResolvePathWithBaseAndRef (dirdb_base, chargen_string, DIRDB_RESOLVE_DRIVE | DIRDB_RESOLVE_TILDE_HOME | DIRDB_RESOLVE_WINDOWS_SLASH, dirdb_use_file);
 #else
 		kernal_ref  = dirdbAPI->ResolvePathWithBaseAndRef (dirdb_base, kernal_string,  DIRDB_RESOLVE_DRIVE | DIRDB_RESOLVE_TILDE_HOME, dirdb_use_file);
 		basic_ref   = dirdbAPI->ResolvePathWithBaseAndRef (dirdb_base, basic_string,   DIRDB_RESOLVE_DRIVE | DIRDB_RESOLVE_TILDE_HOME, dirdb_use_file);
 		chargen_ref = dirdbAPI->ResolvePathWithBaseAndRef (dirdb_base, chargen_string, DIRDB_RESOLVE_DRIVE | DIRDB_RESOLVE_TILDE_HOME, dirdb_use_file);
-
 #endif
 		uint8_t *kernalRom  = loadRom ( kernal_ref, 8192, dirdbAPI);
 		uint8_t *basicRom   = loadRom (  basic_ref, 8192, dirdbAPI);
@@ -280,8 +314,6 @@ namespace libsidplayfp
 		dirdbAPI->Unref (  basic_ref, dirdb_use_file);
 		dirdbAPI->Unref (chargen_ref, dirdb_use_file);
 
-		dirdbAPI->Unref ( dirdb_base, dirdb_use_dir);
-
 		sidplayer.setKernal ( kernalRom);
 		sidplayer.setBasic  (  basicRom);
 		sidplayer.setChargen(chargenRom);
@@ -300,9 +332,8 @@ namespace libsidplayfp
 	uint8_t* ConsolePlayer::loadRom(uint32_t dirdb_ref, const int size, const struct dirdbAPI_t *dirdbAPI)
 	{
 		char *romPath = 0;
-#ifdef __W32__
-		#error we need to make flags, so we can reverse the slashes
-		dirdbAPI->GetFullname_malloc (dirdb_ref, &romPath, DIRDB_FULLNAME_DRIVE);
+#ifdef _WIN32
+		dirdbAPI->GetFullname_malloc (dirdb_ref, &romPath, DIRDB_FULLNAME_DRIVE | DIRDB_FULLNAME_BACKSLASH);
 #else
 		dirdbAPI->GetFullname_malloc (dirdb_ref, &romPath, DIRDB_FULLNAME_NODRIVE);
 #endif
@@ -348,52 +379,30 @@ namespace libsidplayfp
 		// Remove old driver and emulation - should be no-op
 		clearSidEmu ();
 
-		int use_residfp = !strcmp(configAPI->GetProfileString("libsidplayfp", "emulator", "residfp"), "residfp");
+		int use_residfp = !strcmp(configAPI->GetProfileString("libsidplayfp", "emulator", "residfp"), "residfp"); // true for residfp, false for residfpII
 
 		//fprintf (stderr, "use_residfp=%d\n", use_residfp);
 
 		try
 		{
-			if (use_residfp)
+			if (!use_residfp)
 			{
-				ReSIDfpBuilder *rs = new ReSIDfpBuilder( "ReSIDFP" );
-
+				m_residfp = 2;
+				ReSIDfpIIBuilder *rs = new ReSIDfpIIBuilder( "ReSIDFPII" );
 				m_engCfg.sidEmulation = rs;
-				if (!rs->getStatus())
-				{
-					goto createSidEmu_error;
-				}
-
-				rs->create ((sidplayer.info()).maxsids());
-				if (!rs->getStatus())
-				{
-					goto createSidEmu_error;
-				}
-
-				if (m_filter.filterCurve6581)
-				{
-					rs->filter6581Curve(m_filter.filterCurve6581);
-				}
-				if (m_filter.filterCurve8580)
-				{
-					rs->filter8580Curve(m_filter.filterCurve8580);
-				}
+				rs->filter6581Curve(m_filter.filterCurve6581);
+				rs->filter6581Range(m_filter.filterRange6581);
+				rs->filter8580Curve(m_filter.filterCurve8580);
+				rs->combinedWaveformsStrength(m_filter.combinedWaveforms);
 			} else {
-				ReSIDBuilder *rs = new ReSIDBuilder( "ReSID" );
-
+				m_residfp = 1;
+				ReSIDfpBuilder *rs = new ReSIDfpBuilder( "ReSIDFP" );
 				m_engCfg.sidEmulation = rs;
-				if (!rs->getStatus())
-				{
-					goto createSidEmu_error;
-				}
+				rs->filter6581Curve(m_filter.filterCurve6581);
+				rs->filter6581Range(m_filter.filterRange6581);
+				rs->filter8580Curve(m_filter.filterCurve8580);
+				rs->combinedWaveformsStrength(m_filter.combinedWaveforms);
 
-				rs->create ((sidplayer.info ()).maxsids());
-				if (!rs->getStatus())
-				{
-					goto createSidEmu_error;
-				}
-
-				rs->bias(m_filter.bias);
 			}
 		}
 		catch (std::bad_alloc const &ba) {}
@@ -404,7 +413,9 @@ namespace libsidplayfp
 			return false;
 		} else {
 			/* set up SID filter. HardSID just ignores call with def. */
-			m_engCfg.sidEmulation->filter(m_filter.enabled);
+			sidplayer.filter(0, m_filter.enabled);
+			sidplayer.filter(1, m_filter.enabled);
+			sidplayer.filter(2, m_filter.enabled);
 		}
 
 		return true;
@@ -432,7 +443,9 @@ namespace libsidplayfp
 			return false;
 		}
 
-		return open();
+		bool retval = open();
+
+		return retval;
 	}
 
 
@@ -452,6 +465,9 @@ namespace libsidplayfp
 			return false;
 		}
 
+		// First now we know number of SID chips. Since sidplayer.load() resets the buffers, so need to perform initMixer here, and not in load()
+		sidplayer.initMixer(true); // false = mono, true = stereo
+
 		m_state = playerRunning;
 
 		return true;
@@ -467,34 +483,33 @@ namespace libsidplayfp
 			{
 				selected_track = 1;
 			}
-			sidplayer.stop ();
+//			sidplayer.stop ();
 			return open ();
 		}
 		return false;
 	}
 
-	bool ConsolePlayer::iterateaudio (int16_t *targetBuffer, uint_least32_t count, std::vector<int16_t *> *rawSamples)
+
+	uint_least32_t ConsolePlayer::iterateaudio (int16_t *targetBuffer, uint_least32_t count, uint_least32_t cycles, std::vector<int16_t *> *rawSamples)
 	{
 		if (m_state == playerRunning)
 		{
-			if (sidplayer.play (targetBuffer, count*2, rawSamples) < 0)
+			uint_least32_t samples_available = sidplayer.play (cycles);
+			if (samples_available > count)
 			{
-				if (sidplayer.isPlaying ())
-				{
-					m_state = playerError;
-				}
-				return false;
+				fprintf (stderr, "[sidplayer] Warning: sample count %u > samples available %u\n", (unsigned int)samples_available, (unsigned int)count);
+				samples_available = count;
 			}
-			return true;
+			return sidplayer.mix(targetBuffer, samples_available, rawSamples);
 		}
-		return false;
+		return 0;
 	}
 
 	void ConsolePlayer::close (void)
 	{
 		if (m_state != playerStopped)
 		{
-			sidplayer.stop ();
+			//sidplayer.stop ();
 		}
 		m_state = playerStopped;
 		clearSidEmu ();
@@ -513,6 +528,106 @@ namespace libsidplayfp
 		sidplayer.mute (chan / 3, chan % 3, mute);
 	}
 
+	void ConsolePlayer::SetFilter(bool enable)
+	{
+		const SidConfig &config = sidplayer.config ();
+		sidplayer.filter(0, enable);
+		sidplayer.filter(1, enable);
+		sidplayer.filter(2, enable);
+	}
+
+	void ConsolePlayer::SetFilterCurve6581 (double v)
+	{
+		if (v > 1.0) v = 1.0;
+		if (v < 0.0) v = 0.0;
+
+		if (m_residfp == 1)
+		{
+			ReSIDfpBuilder *rs = dynamic_cast<ReSIDfpBuilder *>(m_engCfg.sidEmulation);
+			if (rs)
+			{
+				rs->filter6581Curve (v);
+			}
+		} else if (m_residfp == 2)
+		{
+			ReSIDfpIIBuilder *rs = dynamic_cast<ReSIDfpIIBuilder *>(m_engCfg.sidEmulation);
+			if (rs)
+			{
+				rs->filter6581Curve (v);
+			}
+		}
+	}
+
+	void ConsolePlayer::SetFilterRange6581 (double v)
+	{
+		if (v > 1.0) v = 1.0;
+		if (v < 0.0) v = 0.0;
+
+		if (m_residfp == 1)
+		{
+			ReSIDfpBuilder *rs = dynamic_cast<ReSIDfpBuilder *>(m_engCfg.sidEmulation);
+			if (rs)
+			{
+				rs->filter6581Range (v);
+			}
+		} else if (m_residfp == 2)
+		{
+			ReSIDfpIIBuilder *rs = dynamic_cast<ReSIDfpIIBuilder *>(m_engCfg.sidEmulation);
+			if (rs)
+			{
+				rs->filter6581Range (v);
+			}
+		}
+	}
+
+	void ConsolePlayer::SetFilterCurve8580 (double v)
+	{
+		if (v > 1.0) v = 1.0;
+		if (v < 0.0) v = 0.0;
+
+		if (m_residfp == 1)
+		{
+			ReSIDfpBuilder *rs = dynamic_cast<ReSIDfpBuilder *>(m_engCfg.sidEmulation);
+			if (rs)
+			{
+				rs->filter8580Curve (v);
+			}
+		} else if (m_residfp == 2)
+		{
+			ReSIDfpIIBuilder *rs = dynamic_cast<ReSIDfpIIBuilder *>(m_engCfg.sidEmulation);
+			if (rs)
+			{
+				rs->filter8580Curve (v);
+			}
+		}
+	}
+
+	void ConsolePlayer::SetCombinedWaveformsStrength (int CWS)
+	{
+		if (CWS < 0) CWS = 0;
+		if (CWS > 2) CWS = 2;
+
+		if (m_residfp == 1)
+		{
+			ReSIDfpBuilder *rs = dynamic_cast<ReSIDfpBuilder *>(m_engCfg.sidEmulation);
+			if (rs)
+			{
+				if (CWS == 0) rs->combinedWaveformsStrength (SidConfig::AVERAGE);
+				if (CWS == 1) rs->combinedWaveformsStrength (SidConfig::WEAK);
+				if (CWS == 2) rs->combinedWaveformsStrength (SidConfig::STRONG);
+			}
+		} else if (m_residfp == 2)
+		{
+			ReSIDfpIIBuilder *rs = dynamic_cast<ReSIDfpIIBuilder *>(m_engCfg.sidEmulation);
+			if (rs)
+			{
+				if (CWS == 0) rs->combinedWaveformsStrength (SidConfig::AVERAGE);
+				if (CWS == 1) rs->combinedWaveformsStrength (SidConfig::WEAK);
+				if (CWS == 2) rs->combinedWaveformsStrength (SidConfig::STRONG);
+			}
+		}
+	}
+
 	const SidTuneInfo::clock_t ConsolePlayer::getTuneInfoClockSpeed(void)
 	{
 		const SidTuneInfo *tuneInfo = m_tune.getInfo();
@@ -542,6 +657,13 @@ namespace libsidplayfp
 		return modelData[sidplayer.getModel()].vicModel;
 	}
 
+	const uint32_t ConsolePlayer::GetVICIICyclesPerFrame(void)
+	{
+		c64vic vic = sidplayer.getVIC();
+		return vic.maxRasters *
+		       vic.cyclesPerLine;
+	}
+
 	const char *ConsolePlayer::getCIAmodel(void)
 	{
 		const SidConfig &config = sidplayer.config ();
